import json
import msgpack
import requests


class JsonRpcClient(object):

    codec = json
    mimetype = "application/json"

    def __init__(self, endpoint=None):
        self._id = 1
        self._endpoint = endpoint
        self._session = requests.Session()

    def __getattr__(self, method):
        assert self._endpoint is not None

        def call(**kwargs):
            return self._session.post(
                self._endpoint,
                data=self.codec.dumps(self.request(method, **kwargs)),
                headers={"Content-Type": self.mimetype})
        return call

    def call(self, method, **kwargs):
        return getattr(self, method)(**kwargs)

    def request(self, method, **kwargs):
        req = dict(
            method=method,
            id=str(self._id),
            jsonrpc='2.0',
            params=kwargs,
        )
        return req

    @staticmethod
    def response(resp):
        if 'result' in resp:
            return resp['result']

        if 'error' not in resp:
            raise Exception('invalid JSON-RPC protocol: missing error')

        raise Exception('{}:\n{}'.format(
            resp['error']['message'],
            str(resp['error'].get('data', ''))))


class MsgpackRpcClient(JsonRpcClient):
    codec = msgpack
    mimetype = "application/x-msgpack"



#!/usr/bin/env python
from setuptools import setup
from pymarketstore import __version__

setup(
    name='pymarketstore',
    version=__version__,
    description='Marketstore python driver',
    author='Alpaca',
    author_email='oss@alpaca.markets',
    url='https://github.com/alpacahq/pymarketstore',
    keywords='database,pandas,financial,timeseries',
    packages=['pymarketstore', ],
    install_requires=[
        'msgpack-python',
        'numpy',
        'requests',
        'six',
        'urllib3',
        'pytest',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'coverage>=4.4.1',
        'mock>=1.0.1'
    ],
    setup_requires=['pytest-runner', 'flake8'],
)



import numpy as np
import pandas as pd
import six


def decode(packed):
    dt = np.dtype([
        (colname, coltype)
        for colname, coltype in zip(packed['names'], packed['types'])
    ])
    array = np.empty((packed['length'],), dtype=dt)
    for idx, name in enumerate(dt.names):
        array[name] = np.frombuffer(packed['data'][idx], dtype=dt[idx])
    return array


def decode_responses(responses):
    results = []
    for response in responses:
        packed = response['result']
        array_dict = {}
        array = decode(packed)
        for tbk, start_idx in six.iteritems(packed['startindex']):
            length = packed['lengths'][tbk]
            key = tbk.split(':')[0]
            array_dict[key] = array[start_idx:start_idx + length]
        results.append(array_dict)
    return results


class DataSet(object):

    def __init__(self, array, key, reply):
        self.array = array
        self.key = key
        self.reply = reply

    @property
    def timezone(self):
        return self.reply['timezone']

    @property
    def symbol(self):
        return self.key.split('/')[0]

    @property
    def timeframe(self):
        return self.key.split('/')[1]

    @property
    def attribute_group(self):
        return self.key.split('/')[2]

    def df(self):
        idxname = self.array.dtype.names[0]
        df = pd.DataFrame(self.array).set_index(idxname)
        index = pd.to_datetime(df.index, unit='s', utc=True)
        tz = self.timezone
        if tz.lower() != 'utc':
            index = index.tz_convert(tz)
        df.index = index
        return df

    def __repr__(self):
        a = self.array
        return f'DataSet(key={self.key}, shape={a.shape}, dtype={a.dtype})'


class QueryResult(object):

    def __init__(self, result, reply):
        self.result = {
            key: DataSet(value, key, reply)
            for key, value in six.iteritems(result)
        }
        self.reply = reply

    @property
    def timezone(self):
        return self.reply['timezone']

    def keys(self):
        return list(self.result.keys())

    def first(self):
        return self.result[self.keys()[0]]

    def all(self):
        return self.result

    def __repr__(self):
        content = '\n'.join([
            str(ds) for _, ds in six.iteritems(self.result)
        ])
        return f'QueryResult({content})'


class QueryReply(object):

    def __init__(self, reply):
        results = decode_responses(reply['responses'])
        self.results = [QueryResult(result, reply) for result in results]
        self.reply = reply

    @property
    def timezone(self):
        return self.reply['timezone']

    def first(self):
        return self.results[0].first()

    def all(self):
        datasets = {}
        for result in self.results:
            datasets.update(result.all())
        return datasets

    def keys(self):
        keys = []
        for result in self.results:
            keys += result.keys()
        return keys

    def get_catkeys(self, catnum):
        ret = set()
        for key in self.keys():
            elems = key.split('/')
            ret.add(elems[catnum])
        return list(ret)

    def symbols(self):
        return self.get_catkeys(0)

    def timeframes(self):
        return self.get_catkeys(1)

    def by_symbols(self):
        datasets = self.all()
        ret = {}
        for key, dataset in six.iteritems(datasets):
            symbol = key.split('/')[0]
            ret[symbol] = dataset
        return ret

    def __repr__(self):
        content = '\n'.join([
            str(res) for res in self.results
        ])
        return f'QueryReply({content})'



import numpy as np
import pandas as pd
import requests
import logging

from . import jsonrpc
from .results import QueryReply

logger = logging.getLogger(__name__)


data_type_conv = {
    '<f4': 'f',
    '<f8': 'd',
    '<i4': 'i',
    '<i8': 'q',
}


def isiterable(something):
    return isinstance(something, (list, tuple, set))


def get_rpc_client(codec='msgpack'):
    if codec == 'msgpack':
        return jsonrpc.MsgpackRpcClient
    return jsonrpc.JsonRpcClient


def get_timestamp(value):
    if value is None:
        return None
    if isinstance(value, (int, np.integer)):
        return pd.Timestamp(value, unit='s')
    return pd.Timestamp(value)


class Params(object):

    def __init__(self, symbols, timeframe, attrgroup,
                 start=None, end=None,
                 limit=None, limit_from_start=None):
        if not isiterable(symbols):
            symbols = [symbols]
        self.tbk = ','.join(symbols) + "/" + timeframe + "/" + attrgroup
        self.key_category = None  # server default
        self.start = get_timestamp(start)
        self.end = get_timestamp(end)
        self.limit = limit
        self.limit_from_start = limit_from_start
        self.functions = None

    def set(self, key, val):
        if not hasattr(self, key):
            raise AttributeError()
        if key in ('start', 'end'):
            setattr(self, key, get_timestamp(val))
        else:
            setattr(self, key, val)
        return self

    def __repr__(self):
        content = (f'tbk={self.tbk}, start={self.start}, end={self.end}, ' +
                   f'limit={self.limit}, ' +
                   f'limit_from_start={self.limit_from_start}')
        return f'Params({content})'


class Client(object):

    def __init__(self, endpoint='http://localhost:5993/rpc'):
        self.endpoint = endpoint
        rpc_client = get_rpc_client('msgpack')
        self.rpc = rpc_client(self.endpoint)

    def _request(self, method, **query):
        try:
            resp = self.rpc.call(method, **query)
            resp.raise_for_status()
            rpc_reply = self.rpc.codec.loads(resp.content, encoding='utf-8')
            return self.rpc.response(rpc_reply)
        except requests.exceptions.HTTPError as exc:
            logger.exception(exc)
            raise

    def query(self, params):
        if not isiterable(params):
            params = [params]
        query = self.build_query(params)
        reply = self._request('DataService.Query', **query)
        return QueryReply(reply)

    def write(self, recarray, tbk):
        data = {}
        data['types'] = [
            recarray.dtype[name].str.replace('<', '')
            for name in recarray.dtype.names
        ]
        data['names'] = recarray.dtype.names
        data['data'] = [
            bytes(memoryview(recarray[name]))
            for name in recarray.dtype.names
        ]
        data['length'] = len(recarray)
        data['startindex'] = {tbk: 0}
        data['lengths'] = {tbk: len(recarray)}
        write_request = {}
        write_request['dataset'] = data
        write_request['isvariablelength'] = False
        writer = {}
        writer['requests'] = [write_request]
        try:
            reply = self.rpc.call("DataService.Write", **writer)
        except requests.exceptions.ConnectionError:
            raise requests.exceptions.ConnectionError(
                "Could not contact server")
        reply_obj = self.rpc.codec.loads(reply.content, encoding='utf-8')
        resp = self.rpc.response(reply_obj)
        return resp

    def build_query(self, params):
        reqs = []
        if not isiterable(params):
            params = [params]
        for param in params:
            req = {
                'destination': param.tbk,
            }
            if param.key_category is not None:
                req['key_category'] = param.key_category
            if param.start is not None:
                req['epoch_start'] = int(param.start.value / (10 ** 9))
            if param.end is not None:
                req['epoch_end'] = int(param.end.value / (10 ** 9))
            if param.limit is not None:
                req['limit_record_count'] = int(param.limit)
            if param.limit_from_start is not None:
                req['limit_from_start'] = bool(param.limit_from_start)
            if param.functions is not None:
                req['functions'] = param.functions
            reqs.append(req)
        return {
            'requests': reqs,
        }

    def __repr__(self):
        return f'Client("{self.endpoint}")'
